/* Hekr '48-Passthrough' Protocol
 * Copyright(c) 2018 Hangzhou District Nine Technology Co., Ltd.  */

#include "md5sum.h"
#include "hekr48trans.h"

#define PRODKEY_LEN (32)
#define DEVTID_LEN (32)
#define DEVPRIKEY_LEN (32)
#define RANDOMKEY_LEN (32)

#define CMD_REQUEST_VERIFY '1'
#define CMD_VERIFY '3'
#define CMD_HEARTBEAT 'B'
#define CMD_APPSENDRESP '8'
#define CMD_DEVSEND '9'

static ra_size_t _strlen(const char *s)
{
    ra_size_t len = 0;
    const char *p = s;
    while (*p++ != '\0') len++;
    return len;
}

static void *_memcpy(void *dest, const void *src, ra_size_t n)
{
    char *dest_p = dest;
    const char *src_p = src;
    while (n-- != 0)
    {
        *dest_p++ = *src_p++;
    }
    return dest;
}

/* 0x0A -> "0A", by 2 */
int raw_to_hexlit(char *dst, const ra_u8 *src, ra_size_t len)
{
    const ra_u8 *src_p = src;
    char *dst_p = dst;

    if (len == 0) return -1;

    while (len-- != 0)
    {
#define _x(_s) \
        do { *dst_p++ = "0123456789ABCDEF"[_s]; } while (0)

        _x((*src_p) >> 4);
        _x((*src_p) & 0xF);
        src_p++;

#undef _x
    }
    *dst_p = '\0';

    return (int)(dst_p - dst);
}

/* "0A" -> 0x0A, by 2 */
int hexlit_to_raw(ra_u8 *dst, const char *src, const ra_size_t len)
{
    ra_size_t i;
    ra_u8 *dst_p = dst;
    const char *src_p = src;

    if ((len & 1) != 0) { return -1; }

    for (i = 0; i != len; i += 2)
    {
        ra_u8 t_hi, t_lo;

#define _x(_d, _s) \
        do { \
            if ('0' <= _s && _s <= '9') { _d = (ra_u8)(_s - '0'); } \
            else if ('a' <= _s && _s <= 'f') { _d = (ra_u8)(_s - 'a' + 10); } \
            else if ('A' <= _s && _s <= 'F') { _d = (ra_u8)(_s - 'A' + 10); } \
            else return -1; \
        } while (0)
        _x(t_hi, *src_p); src_p++;
        _x(t_lo, *src_p); src_p++;
#undef _x

        *dst_p++ = (ra_u8)((t_hi << 4) | (t_lo));
    }

    return (int)(dst_p - dst);
}

int hexlit_to_raw16(ra_u16 *dst, const char *src, const ra_size_t len)
{
    ra_u8 hi, lo;

    if (len != 4) return -1;

    hexlit_to_raw(&hi, src, 2);
    hexlit_to_raw(&lo, src + 2, 2);
    *dst = ((ra_u16)hi << 8) | ((ra_u16)lo);

    return 0;
}

static ra_u8 checksum8(ra_u8 *data, ra_size_t len)
{
    ra_u32 cs = 0;
    ra_u8 *p = data;

    while (len != 0)
    {
        ra_u8 t;
        if (hexlit_to_raw(&t, (const char *)p, 2) < 0)
        {
            /* Error */
            return 0;
        }
        cs += t;

        p += 2;
        len -= 2;
    }

    return cs & 0xFF;
}

int h48_build_frame_verify_request(ra_u8 *buf, const ra_size_t buf_size, \
        const ra_u8 frame_num, const char *prodkey, const char *devtid)
{
    if (_strlen(prodkey) != PRODKEY_LEN) return -1;
    if (_strlen(devtid) != DEVTID_LEN) return -1;
    if (buf_size < 8 + PRODKEY_LEN * 2 + DEVTID_LEN * 2 + 2) return -1;

    buf[0] = '4'; buf[1] = '8'; /* 48 */
    buf[2] = '4'; buf[3] = '5'; /* length */
    buf[4] = '0'; buf[5] = CMD_REQUEST_VERIFY; /* command */
    raw_to_hexlit((char *)buf + 6, &frame_num, 1); /* frame number */
    raw_to_hexlit((char *)buf + 8, (const ra_u8 *)prodkey, (ra_size_t)_strlen(prodkey)); /* prodkey */
    raw_to_hexlit((char *)buf + 8 + PRODKEY_LEN * 2, (const ra_u8 *)devtid, (ra_size_t)_strlen(devtid)); /* prodkey */
    {
        ra_u8 cs = checksum8(buf, 8 + PRODKEY_LEN * 2 + DEVTID_LEN * 2);
        raw_to_hexlit((char *)buf + 8 + PRODKEY_LEN * 2 + DEVTID_LEN * 2, &cs, 1);
    }

    return 8 + PRODKEY_LEN * 2 + DEVTID_LEN * 2 + 2;
}

int h48_build_frame_verify(ra_u8 *buf, const ra_size_t buf_size, \
        const ra_u8 frame_num, const char *randomkey, const char *devtid, const char *devprikey)
{
    if (_strlen(randomkey) != RANDOMKEY_LEN) return -1;
    if (_strlen(devtid) != DEVTID_LEN) return -1;
    if (_strlen(devprikey) != DEVPRIKEY_LEN) return -1;

    if (buf_size < 8 + 16 * 2 + 2) return -1;

#if 0
    printf("randomkey: %s\n", randomkey);
    printf("devtid: %s\n", devtid);
    printf("devprikey: %s\n", devprikey);
#endif

    buf[0] = '4'; buf[1] = '8'; /* 48 */
    buf[2] = '1'; buf[3] = '5'; /* length */
    buf[4] = '0'; buf[5] = CMD_VERIFY; /* command */
    raw_to_hexlit((char *)buf + 6, &frame_num, 1); /* frame number */
    {
        /* Compute AuthKey */
        MD5_CTX context;
        unsigned char md5_digest[16];

        MD5Init(&context);
        MD5Update(&context, (unsigned char *)randomkey, (ra_size_t)RANDOMKEY_LEN);
        MD5Update(&context, (unsigned char *)devtid, (ra_size_t)DEVTID_LEN);
        MD5Update(&context, (unsigned char *)devprikey, (ra_size_t)DEVPRIKEY_LEN);
        MD5Final(&context, md5_digest);
        raw_to_hexlit((char *)buf + 8, md5_digest, 16);

#if 0
        printf("authkey: ");
        fwrite(buf + 8, 32, 1, stdout);
        printf("\n");
#endif
    }
    {
        ra_u8 cs = checksum8(buf, 8 + 16 * 2);
        raw_to_hexlit((char *)buf + 8 + 16 * 2, &cs, 1);
    }

    return 8 + 16 * 2 + 2;
}

int h48_build_frame_heartbeat(ra_u8 *buf, const ra_size_t buf_size, \
        const ra_u8 frame_num)
{
    if (buf_size < 8 + 2) return -1;

    buf[0] = '4'; buf[1] = '8'; /* 48 */
    buf[2] = '0'; buf[3] = '5'; /* length */
    buf[4] = '0'; buf[5] = CMD_HEARTBEAT; /* command */
    raw_to_hexlit((char *)buf + 6, &frame_num, 1); /* frame number */
    {
        ra_u8 cs = checksum8(buf, 8);
        raw_to_hexlit((char *)buf + 8, &cs, 1);
    }

    return 8 + 2;
}

int h48_build_frame_devsend(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num, const ra_u16 msgid, \
        const ra_u8 *data, const ra_size_t data_len)
{
    if (buf_size < 12 + data_len + 2) return -1;

    buf[0] = '4'; buf[1] = '8'; /* 48 */
    {
        ra_u8 frame_len = (ra_u8)(6 + (data_len >> 1)) + 1;
        raw_to_hexlit((char *)buf + 2, &frame_len, 1); /* length */
    }
    buf[4] = '0'; buf[5] = CMD_DEVSEND; /* command */
    raw_to_hexlit((char *)buf + 6, &frame_num, 1); /* frame number */
    {
        ra_u8 msgid_hi = (ra_u8)(msgid >> 8);
        ra_u8 msgid_lo = (ra_u8)(msgid & 0xFF);
        raw_to_hexlit((char *)buf + 8, &msgid_hi, 1); /* msgid (hi) */
        raw_to_hexlit((char *)buf + 10, &msgid_lo, 1); /* msgid (lo) */
    }
    _memcpy(buf + 12, data, data_len); /* data */
    {
        ra_u8 cs = checksum8(buf, 12 + data_len);
        raw_to_hexlit((char *)buf + 12 + data_len, &cs, 1);
    }

    return (int)(12 + data_len + 2);
}

int h48_build_frame_appsendresp(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num, const ra_u16 msgid, \
        const ra_u8 *apptid)
{
    if (buf_size < 74 * 2 + 2) return -1;

    buf[0] = '4'; buf[1] = '8'; /* 48 */
    buf[2] = '4'; buf[3] = 'B'; /* length */
    buf[4] = '0'; buf[5] = CMD_APPSENDRESP; /* command */
    raw_to_hexlit((char *)buf + 6, &frame_num, 1); /* frame number */
    {
        ra_u8 msgid_hi = (ra_u8)(msgid >> 8);
        ra_u8 msgid_lo = (ra_u8)(msgid & 0xFF);
        raw_to_hexlit((char *)buf + 8, &msgid_hi, 1); /* msgid (hi) */
        raw_to_hexlit((char *)buf + 10, &msgid_lo, 1); /* msgid (lo) */
    }
    _memcpy(buf + 12, apptid, 64 * 2); /* apptid */
    /* reserved */
    {
        int i;
        for (i = 0; i != 8; i++)
        {
            buf[12 + 64 * 2 + i] = '0';
        }
    }
    {
        ra_u8 cs = checksum8(buf, 12 + 64 * 2 + 8);
        raw_to_hexlit((char *)buf + 12 + 64 * 2 + 8, &cs, 1);
    }

    return 12 + 64 * 2 + 8 + 2;
}

h48_frame_type h48_frame_decode_type(ra_u8 *frame, const ra_size_t frame_size)
{
    h48_frame_type ret;
    ra_u8 v;
    if (frame_size < 7) return H48_FRAME_TYPE_UNKNOWN;

    hexlit_to_raw(&v, (const char *)frame + 4, 2);
    switch (v)
    {
        case 0x07: ret = H48_FRAME_TYPE_APPSEND; break;
        default: ret = H48_FRAME_TYPE_UNKNOWN; break;
    }

    return ret;
}

