#include <sys/time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <string.h>
#include <stdio.h>

#include "ra_types.h"
#include "usocket.h"


/* Initialize usocket */
void usocket_init( \
        usocket_t *usocket)
{
    usocket->type = USOCKET_TYPE_NONE;
}

/* If the fd is writable */
static int _writable(int fd, const int timeout)
{
    fd_set wfds;
    struct timeval tv;
    int retval;

    FD_ZERO(&wfds);
    FD_SET(fd, &wfds);

    tv.tv_sec = timeout;
    tv.tv_usec = 0;

    retval = select(fd + 1, NULL, &wfds, NULL, &tv);
    if (retval == -1) return -1;
    else if (retval == 0) return 0;
    /* retval >= 0 */

    return 1;
}

/* Connect to a plain host
 * @return 0  success
 * @return -1 fail */
int usocket_connect3( \
        usocket_t *usocket, \
        const char *host, const ra_size_t port, const int timeout, int blocking)
{
    int ret = 0;
    int fd = -1;
    /* struct hostent *hent; */
    struct sockaddr_in addr;
    /* char *dst_p, *src_p; */
    struct addrinfo *servinfo = NULL;

    if (usocket->type != USOCKET_TYPE_NONE) { return -1; }

    /* Resolve the host */
    struct addrinfo hints;
    char service_buf[16 + 1];
    snprintf(service_buf, 16, "%d", (int)port);
    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;
    if (getaddrinfo(host, service_buf, &hints, &servinfo) != 0)
    { ret = -1; goto fail; }

    /* Fill address */
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons((uint16_t)port);
    memcpy(&addr.sin_addr, &((struct sockaddr_in *)servinfo->ai_addr)->sin_addr, sizeof(struct in_addr));

    /* Create socket */
    if ((fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
    { ret = -1; goto fail; }

    if (timeout == 0)
    {
        /* No timeout */

        /* Connect */
        if (connect(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) < 0)
        { ret = -1; goto fail; }

        /* Non-block */
        if (blocking == 0)
        {
            fcntl(fd, F_SETFL, O_NONBLOCK);
        }
    }
    else
    {
        /* Non-block */
        fcntl(fd, F_SETFL, O_NONBLOCK);

        /* Connect */
        if (connect(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) < 0)
        {

            if (errno != EINPROGRESS) { ret = -1; goto fail; }
            if (!(_writable(fd, timeout) == 1)) { ret = -1; goto fail; }
            {
                int err;
                socklen_t len = sizeof(int);
                if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, &len) < 0)
                { ret = -1; goto fail; }
                if (err != 0)
                { ret = -1; goto fail; }
            }
        }

        /* Non-block */
        if (blocking != 0)
        {
            const int flags = fcntl(fd, F_GETFL, 0);
            fcntl(fd, F_SETFL, flags & (~O_NONBLOCK));
        }
    }

    usocket->type = USOCKET_TYPE_PLAIN;
    usocket->u.as_plain.fd = fd;

    goto done;
fail:
    if (fd != -1) close(fd);
done:
    if (servinfo != NULL)
    {
        freeaddrinfo(servinfo);
    }
    return ret;
}

/* Connect to a host */
int usocket_connect2( \
        usocket_t *usocket, \
        const char *host, const ra_size_t port, const int timeout)
{
    return usocket_connect3(usocket, host, port, timeout, 1);
}

int usocket_connect( \
        usocket_t *usocket, \
        const char *host, const ra_size_t port)
{
    return usocket_connect2(usocket, host, port, 0);
}

/* If the connection is alive */ 
int usocket_alive( \
        usocket_t *usocket)
{
    return usocket->type == USOCKET_TYPE_NONE ? 0 : 1;
}

/* Close */
void usocket_close( \
        usocket_t *usocket)
{
    switch (usocket->type)
    {
        case USOCKET_TYPE_NONE:
            break;

        case USOCKET_TYPE_PLAIN:
            if (usocket->u.as_plain.fd != -1) close(usocket->u.as_plain.fd);
            usocket->u.as_plain.fd = -1;
            usocket->type = USOCKET_TYPE_NONE;
            break;
    }
}


/* If the fd is readable */
static int _readable(int fd)
{
    fd_set rfds;
    struct timeval tv;
    int retval;

    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);

    tv.tv_sec = 0;
    tv.tv_usec = 1;

    retval = select(fd + 1, &rfds, NULL, NULL, &tv);
    if (retval == -1) return -1;
    else if (retval) return 1;

    return 0;
}

/* Readable */
int usocket_readable( \
        usocket_t *usocket)
{
    int ret = 0;

    switch (usocket->type)
    {
        case USOCKET_TYPE_NONE:
            ret = -1;
            break;

        case USOCKET_TYPE_PLAIN:
            ret = _readable(usocket->u.as_plain.fd);
            break;
    }

    return ret;
}

/* Recv */
int usocket_recv( \
        usocket_t *usocket, \
        void *buf, ra_size_t len)
{
    int ret = 0;

    switch (usocket->type)
    {
        case USOCKET_TYPE_NONE:
            ret = -1;
            break;

        case USOCKET_TYPE_PLAIN:
            ret = (int)recv(usocket->u.as_plain.fd, buf, len, 0);
            break;
    }

    return ret;
}

/* Send */
int usocket_send( \
        usocket_t *usocket, \
        const void *buf, ra_size_t len)
{
    int ret = 0;

    switch (usocket->type)
    {
        case USOCKET_TYPE_NONE:
            ret = -1;
            break;

        case USOCKET_TYPE_PLAIN:
            ret = (int)send(usocket->u.as_plain.fd, buf, len, MSG_NOSIGNAL);
            break;
    }

    return ret;
}

